"""Marshmallow schemas for model definitions about an academic institution.

Autogenerated with marshmallow-mongoengine.
"""
from gator._vendor.marshmallow_mongoengine import ModelSchema

import gator.core.models.institution as institution


class InstitutionSchema(ModelSchema):
    """Marshmallow schema for the :class:`Institution` model."""

    class Meta:
        """Meta class for InstitutionSchema."""

        model = institution.Institution


class BuildingSchema(ModelSchema):
    """Marshmallow schema for the :class:`Building` model."""

    class Meta:
        """Meta class for BuildingSchema."""

        model = institution.Building


class LocationSchema(ModelSchema):
    """Marshmallow schema for the :class:`Location` model."""

    class Meta:
        """Meta class for LocationSchema."""

        model = institution.Location
